#include "nrf.h"
#include "bsp.h"
#include "ff.h"
#include "diskio_blkdev.h"
#include "nrf_block_dev_sdc.h"

#include <stdbool.h>                    //UART libraries
#include <stdint.h>
#include <stdio.h>
#include "app_uart.h"
#include "app_error.h"
#include "nrf_delay.h"


#include "nrf_gpio.h"                   //RTC Libraries
#include "nrf_drv_rtc.h"
#include "nrf_drv_clock.h"
#include "boards.h"

#include "RTcounter.h"


//#define NRF_LOG_MODULE_NAME "APP"
//#include "nrf_log.h"
//#include "nrf_log_ctrl.h"

//========================================RTC Start==============================================================================

#define COMPARE_COUNTERTIME  (3UL)                                        /**< Get Compare event COMPARE_TIME seconds after the counter starts from 0. */

#ifdef BSP_LED_0
    #define TICK_EVENT_OUTPUT     BSP_LED_0                                 /**< Pin number for indicating tick event. */
#endif
#ifndef TICK_EVENT_OUTPUT
    #error "Please indicate output pin"
#endif
#ifdef BSP_LED_1
    #define COMPARE_EVENT_OUTPUT   BSP_LED_1                                /**< Pin number for indicating compare event. */
#endif
#ifndef COMPARE_EVENT_OUTPUT
    #error "Please indicate output pin"
#endif

const nrf_drv_rtc_t rtc = NRF_DRV_RTC_INSTANCE(2); /**< Declaring an instance of nrf_drv_rtc for RTC0. */

extern bool ov_flag;
extern bool time_req_flag, set_date_time_flag;

time t,p;

/** @brief Function starting the internal LFCLK XTAL oscillator.
 */
static void lfclk_config(void)
{
    ret_code_t err_code = nrf_drv_clock_init();
    APP_ERROR_CHECK(err_code);

    nrf_drv_clock_lfclk_request(NULL);
}

/** @brief Function initialization and configuration of RTC driver instance.
 */

static void rtc_handler(nrf_drv_rtc_int_type_t int_type)
{
    if (int_type == NRF_DRV_RTC_INT_COMPARE0)
    {

    }
    else if (int_type == NRF_DRV_RTC_INT_TICK)
    {
        //nrf_gpio_pin_toggle(TICK_EVENT_OUTPUT);
    }
    else if (int_type == NRF_DRV_RTC_INT_OVERFLOW)
    {
        ov_flag=true;
        get_time_date();
        memcpy(&t,&p,sizeof(p));	// memcpy (dest_struct, source_struct, sizeof (dest_struct));
        printf("DATE_TIME after overflow\r\nDATE =  %d-%d-%d TIME = %d:%d:%d\r\n",t.date,t.month,t.year,t.hour,t.minute,t.second);
    }
}


static void rtc_config(void)
{
    uint32_t err_code;
     //Initialize RTC instance
    nrf_drv_rtc_config_t config = NRF_DRV_RTC_DEFAULT_CONFIG;
    config.prescaler = 0;
    err_code = nrf_drv_rtc_init(&rtc, &config, rtc_handler);
    APP_ERROR_CHECK(err_code);

    //Enable tick event & interrupt
    //nrf_drv_rtc_tick_enable(&rtc,true);
	
    nrf_drv_rtc_overflow_enable(&rtc,true);
    
    //nrf_drv_rtc_enable(&rtc); //RTC is enabled in set_date_time() function
}


//==================================================RTC END==========================================================


//********************************************UART Start***********************************
//*****************************************************************************************
#define MAX_TEST_DATA_BYTES     (15U)                /**< max number of test bytes to be used for tx and rx. */
#define UART_TX_BUF_SIZE 256                         /**< UART TX buffer size. */
#define UART_RX_BUF_SIZE 256                         /**< UART RX buffer size. */


void uart_error_handle(app_uart_evt_t * p_event)
{
    if (p_event->evt_type == APP_UART_COMMUNICATION_ERROR)
    {
        APP_ERROR_HANDLER(p_event->data.error_communication);
    }
    else if (p_event->evt_type == APP_UART_FIFO_ERROR)
    {
        APP_ERROR_HANDLER(p_event->data.error_code);
    }
}

void uart_init()
{
    uint32_t err_code;
    const app_uart_comm_params_t comm_params =
          {
              RX_PIN_NUMBER,
              TX_PIN_NUMBER,
              RTS_PIN_NUMBER,
              CTS_PIN_NUMBER,
              APP_UART_FLOW_CONTROL_ENABLED,
              false,
              UART_BAUDRATE_BAUDRATE_Baud250000
          };

        APP_UART_FIFO_INIT(&comm_params,
                             UART_RX_BUF_SIZE,
                             UART_TX_BUF_SIZE,
                             uart_error_handle,
                             APP_IRQ_PRIORITY_LOWEST,
                             err_code);

        APP_ERROR_CHECK(err_code);
}
//********************************************UART END***********************************
//*****************************************************************************************



#define FILE_NAME   "NORDIC.TXT"
#define foldername  "/HTIC"
#define testname  "/HTIC/jam.txt"

#define SDC_SCK_PIN     ARDUINO_13_PIN  ///< SDC serial clock (SCK) pin.
#define SDC_MOSI_PIN    ARDUINO_11_PIN  ///< SDC serial data in (DI) pin.
#define SDC_MISO_PIN    ARDUINO_12_PIN  ///< SDC serial data out (DO) pin.
#define SDC_CS_PIN      ARDUINO_10_PIN  ///< SDC chip select (CS) pin.

static FATFS fs;
static DIR dir;
static FILINFO fno;
static FIL file;
FRESULT ff_result;

uint32_t bytes_written;
int n,eof=0;
TCHAR fileName[13],dirname[13] = {0};

/**
 * @brief  SDC block device definition
 * */
NRF_BLOCK_DEV_SDC_DEFINE(
        m_block_dev_sdc,
        NRF_BLOCK_DEV_SDC_CONFIG(
                SDC_SECTOR_SIZE,
                APP_SDCARD_CONFIG(SDC_MOSI_PIN, SDC_MISO_PIN, SDC_SCK_PIN, SDC_CS_PIN)
         ),
         NFR_BLOCK_DEV_INFO_CONFIG("Nordic", "SDC", "1.00")
);

/**
 * @brief Function for demonstrating FAFTS usage.
 */
static void fatfs_example()
{
    DSTATUS disk_state = STA_NOINIT;

    // Initialize FATFS disk I/O interface by providing the block device.
    static diskio_blkdev_t drives[] =
    {
            DISKIO_BLOCKDEV_CONFIG(NRF_BLOCKDEV_BASE_ADDR(m_block_dev_sdc, block_dev), NULL)
    };

    diskio_blockdev_register(drives, ARRAY_SIZE(drives));

    printf("Initializing disk 0 (SDC)...\r\n");
    for (uint32_t retries = 3; retries && disk_state; --retries)
    {
        disk_state = disk_initialize(0);
    }
    if (disk_state)
    {
        printf("Disk initialization failed.\r\n");
        return;
    }
    
    uint32_t blocks_per_mb = (1024uL * 1024uL) / m_block_dev_sdc.block_dev.p_ops->geometry(&m_block_dev_sdc.block_dev)->blk_size;
    uint32_t capacity = m_block_dev_sdc.block_dev.p_ops->geometry(&m_block_dev_sdc.block_dev)->blk_count / blocks_per_mb;
    printf("Capacity: %d MB\r\n", capacity);

    printf("Mounting volume...\r\n");
    ff_result = f_mount(&fs, "", 1);
    if (ff_result)
    {
        printf("Mount failed.\r\n");
        return;
    }
    return;
}


//***************************************************
//Function to transmit a single byte
//***************************************************
void transmitByte( unsigned char data )
{
    while (app_uart_put(data) != NRF_SUCCESS);
}

//**************************************************
//Function to receive a single byte
//*************************************************
unsigned char receiveByte( void )
{
    unsigned char data;
    while (app_uart_get(&data) != NRF_SUCCESS);

    return(data);
}

void get_name()
{
    memset(fileName,0,sizeof(fileName));
    unsigned char data = 0;
    uint8_t ii=0;
    
    printf("Enter the name of file/Directory...\r\n");
    
    while(1)
    {
        data = receiveByte();
        transmitByte(data);
        
        if(data == 0x0d) 
        {
            (fileName[ii]) = '\0';
            break;  //'ENTER' key pressed
        }
        
        (fileName[ii]) = data;
        ii++; 
    }
    printf("entered string is = %s\r\n",fileName);
}


void sd_append_file()
{
    char buffer [100]={0};
    uint8_t d[20] = {255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255};
    
    get_name();             // Get file name from User
    
    ff_result = f_open(&file, fileName, FA_WRITE | FA_OPEN_ALWAYS);
    if (ff_result == FR_OK) {
        /* Seek to end of the file to append data */
        ff_result = f_lseek(&file, f_size(&file));
        if (ff_result != FR_OK)
            f_close(&file);
    }
    
   n=sprintf (buffer,"%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\r",d[0],d[1],d[2],d[3],d[4],d[5],d[6],d[7],d[8],d[9],d[10],d[11],d[12],d[13],d[14],d[15],d[16],d[17],d[18],d[19]);
   ff_result = f_write(&file, buffer, n, (UINT *) &bytes_written);
    if (ff_result != FR_OK)
    {
        printf("Write failed\r\n.");
        return;
    }
    else
    {
        printf("Write Successful...\r\n.");
    } 
    
    /* Append a line */
//    f_printf(&file, "%02u/%02u/%u, %2u:%02u\r\n");

    /* Close the file */
    f_close(&file);
}

void sd_test()
{
    get_name();

    printf("Opening directory: %s\r\n",fileName);
    ff_result = f_opendir(&dir, fileName);
    if (ff_result != FR_OK)
    {
      printf("Directory not opened\r\n.");
    }
    else
    {
        printf("open DIR successful = %s\r\n",fileName);
    }
    
}

//void get_dir()
//{
//    printf("Enter the DIR name\r\n");
//    get_name();
//    memset(dirname,0,sizeof(dirname));
//    
//    if(fileName[0]!= '/')
//    {
//        memcpy(&dirname[1],&fileName[0],(sizeof(fileName)-1));
//        dirname[0] = '/';
//    }
//    else
//    {
//        memcpy(dirname,fileName,sizeof(fileName));
//    }
//    
//    printf("\r\n Opening directory: /\r\n");
//    ff_result = f_opendir(&dir, dirname);
//    if (ff_result != FR_OK)
//    {
//        printf("Unable to Open DIR\r\n");
//        return;
//    }
//    
//    
//
//}

void sd_create_file()
{    
    get_name();   
    
    ff_result = f_open(&file,fileName, FA_READ | FA_WRITE | FA_OPEN_APPEND);
    if (ff_result != FR_OK)
    {
        printf("Unable to open or create file: %s\r\n",fileName);
        return;
    }
    if (ff_result == FR_OK)
    {
        printf("File Created: %s\r\n",fileName);
    }
}

void  sd_file_list()
{
    TCHAR temp_dir[13] = {0};
    
    get_name();
    if(fileName[0]!= '/')
    {
        memcpy(&temp_dir[1],&fileName[0],(sizeof(fileName)-1));
        temp_dir[0] = '/';
    }
    else
    {
        memcpy(temp_dir,fileName,sizeof(fileName));
    }
    
    printf("\r\n Opening directory: /\r\n");
    ff_result = f_opendir(&dir, temp_dir);
    if (ff_result != FR_OK)
    {
        printf("Unable to Open DIR\r\n");
        return;
    }
    do
    {
        ff_result = f_readdir(&dir, &fno);
        if (ff_result != FR_OK)
        {
            printf("Directory read failed.");
            return;
        }
        
        if (fno.fname[0])
        {
            if (fno.fattrib & AM_DIR)
            {
                printf("   <DIR>   %s\r\n",fno.fname);
            }
            else
            {
                printf("%9lu  %s\r\n", fno.fsize, fno.fname);
            }
        }
    }
    while (fno.fname[0]);
    
    ff_result = f_closedir (&dir);
    if (ff_result != FR_OK)
    {
        printf("Unable to Close DIR\r\n");
        return;
    }
    
    
}

void sd_filesize()
{
    FATFS *fs1;
    DWORD fre_clust, fre_sect, tot_sect;
    
    
    /* Get volume information and free clusters of drive 1 */
    ff_result = f_getfree("", &fre_clust, &fs1);
    if (ff_result != FR_OK)
    {
        printf("Unable to get volume information...\r\n");
    }
    
    /* Get total sectors and free sectors */
    tot_sect = (fs1->n_fatent - 2) * fs1->csize;
    fre_sect = fre_clust * fs1->csize;

    /* Print the free space (assuming 512 bytes/sector) */
    printf("%10lu KiB total drive space.\r\n%10lu KiB available.\r\n", tot_sect / 2, fre_sect / 2);

}

void sd_format()
{
    BYTE work[_MAX_SS]; /* Work area (larger is better for processing time) */
    
    /* Create FAT volume */
    ff_result = f_mkfs("", FM_FAT32, 0, work, sizeof work);
    if (ff_result != FR_OK)
    {
        printf("Unable to format Disk\r\n");
        return;
    }
    if (ff_result == FR_OK)
    {
        printf("Format Complete...\r\n");
        return;
    }
}

void sd_file_remove()
{
    get_name();
  
    //remove perticular file 
    ff_result = f_unlink (fileName);
    if (ff_result != FR_OK)
    {
      printf("NoFile : Write failed\r\n.");
    }
    else
    {
        printf("file removed = ANK.TXT \r\n");
    }
}

void sd_create_dir()
{
    get_name();                 //Get DIR name from USER
    
    ff_result = f_mkdir(fileName);
    if (ff_result != FR_OK)
    {
      printf("Directory not created\r\n.");
    }
    else
    {
        printf("create DIR successful = %s\r\n",fileName);
    }
}

void sd_time_date()
{  
    uint8_t val[3],buff[5]={0},i;
    char buffer [12]={0};
    char temp[2]={0};//temp2=0;
  
    printf("Enter current date and time in DDMMYY and HHMMSS\r\n");
    
    TCHAR fileName[13]={0};
    unsigned char data;
    uint8_t ii=0;
    
    while(1)
    {
        data = receiveByte();
        transmitByte(data);
        
        if(data == 0x0d) 
        {
            (fileName[ii]) = '\0';
            break;  //'ENTER' key pressed
        }
        
        (fileName[ii]) = data;
        ii++; 
    }
    printf("entered string is = %s\r\n",fileName);
    
    for(i=0;i<6;i+=2)
    {
        temp[0] = fileName[i];
        temp[1] = fileName[i+1];
        val[i/2] =(atoi(temp));
    }
    
    printf("%x %x %x \r\n",val[0],val[1],val[2]);
    
    buff[0] = val[0];
    buff[1] = 0x2D;
    buff[2] = val[1];
    buff[3] = 0x2D;
    buff[4] = val[2];
    
    n=sprintf (buffer,"%d%c%d%c%d",buff[0],buff[1],buff[2],buff[3],buff[4]);
   
    
    printf("Timestamp: %u/%02u/%02u, %02u:%02u\r\n",
    (fno.fdate >> 9) + 1980, fno.fdate >> 5 & 15, fno.fdate & 31,
    fno.ftime >> 11, fno.ftime >> 5 & 63);

    printf("Attributes: %c%c%c%c%c\r\n",
    (fno.fattrib & AM_DIR) ? 'D' : '-',
    (fno.fattrib & AM_RDO) ? 'R' : '-',
    (fno.fattrib & AM_HID) ? 'H' : '-',
    (fno.fattrib & AM_SYS) ? 'S' : '-',
    (fno.fattrib & AM_ARC) ? 'A' : '-');
    
    ff_result = f_mkdir(buffer);
    if (ff_result != FR_OK)
    {
      printf("Directory not created\r\n.");
    }
    else
    {
        printf("create DIR successful = %s\r\n",buffer);
    }
}

void create_folder()
{
    uint8_t date_buff[6];
//    uint8_t time_buff[4];
    char buffer [20]={0};
    printf("Final Date and Time\r\n");
    printf("DATE =  %d-%d-%d TIME = %d:%d:%d\r\n",p.date,p.month,p.year,p.hour,p.minute,p.second);
    
    date_buff[0] = p.hour;
    date_buff[1] = 0x2D;
    date_buff[2] = p.minute;
    date_buff[3] = 0x2D;
//    date_buff[4] = p.second;
    date_buff[4] = (uint8_t)((p.year & 0xFF00)>>8);
    date_buff[5] = (uint8_t)(p.year & 0x00FF);
    
    printf("%d%c%d%c%x%x\r\n",date_buff[0],date_buff[1],date_buff[2],date_buff[3],date_buff[4],date_buff[5]);
    n = sprintf(buffer,"%d%c%d%c%d\r\n",date_buff[0],date_buff[1],date_buff[2],date_buff[3],p.year);
    
//    printf("%d%c%d%c%d\r\n",date_buff[0],date_buff[1],date_buff[2],date_buff[3],date_buff[4]);
//    n = sprintf(buffer,"%d%c%d%c%d\r\n",date_buff[0],date_buff[1],date_buff[2],date_buff[3],date_buff[4]);
//    
   
    ff_result = f_stat(buffer, &fno);
   
    switch (ff_result) 
    {
        case FR_OK:
                printf("FOLDER exist..\r\n");
            break;

        case FR_NO_FILE:
            printf("FOLDER is not exist.\r\n");
            ff_result = f_mkdir(buffer);
            if (ff_result != FR_OK)
            {
              printf("Directory not created\r\n.");
            }
            else
            {
                printf("create DIR successful = %s\r\n",buffer);
            }
            break;

        default:
            printf("An error occured. (%d)\n", ff_result);
    }
    
    
    
}


void set_date_time()
{
    printf("Enter the date and time in ddmmyyyyhhmmss format\r\n");
    
    TCHAR fileName[13]={0};
    unsigned char data;
    uint8_t ii=0;
    
    while(1)
    {
        data = receiveByte();
        transmitByte(data);
        
        if(data == 0x0d) 
        {
            (fileName[ii]) = '\0';
            break;  //'ENTER' key pressed
        }
        
        (fileName[ii]) = data;
        ii++; 
    }
    printf("entered string is = %s\r\n",fileName);
    
    uint8_t val[7],mul_fact=0x64;
    //char buffer [12]={0};
    char temp[2]={0};//temp2=0;
    
    for(ii=0;ii<14;ii+=2)
    {
        temp[0] = fileName[ii];
        temp[1] = fileName[ii+1];
        val[ii/2] =(atoi(temp));
    }
    
    printf("%x %x %x %x %x %x %x\r\n",val[0],val[1],val[2],val[3],val[4],val[5],val[6]);
    
    t.date 	= val[0];
    t.month     = val[1];
    t.year 	= (val[2] * mul_fact) + val[3];
    t.hour	= val[4];
    t.minute    = val[5];
    t.second    = val[6];
    
    printf("START DATE\r\nDATE =  %d-%d-%d TIME = %d:%d:%d\r\n",t.date,t.month,t.year,t.hour,t.minute,t.second);
    nrf_drv_rtc_enable(&rtc);
}

void file_list()
{
    printf("\r\n Opening directory: /\r\n");
    ff_result = f_opendir(&dir, foldername);
    if (ff_result)
    {
        printf("Directory listing failed!\r\n");
        return;
    }
    do
    {
        ff_result = f_readdir(&dir, &fno);
        if (ff_result != FR_OK)
        {
            printf("Directory read failed.");
            return;
        }
        
        if (fno.fname[0])
        {
            if (fno.fattrib & AM_DIR)
            {
                printf("   <DIR>   %s\r\n",fno.fname);
            }
            else
            {
                printf("%9lu  %s\r\n", fno.fsize, fno.fname);
            }
        }
    }
    while (fno.fname[0]);
}

void read_dir()
{
    printf("Enter the name of DIR to be Read..\r\n");
    memset(fileName,0,sizeof(fileName));
    unsigned char data = 0;
    uint8_t ii=0;
    
    while(1)
    {
        data = receiveByte();
        transmitByte(data);
        
        if(data == 0x0d) 
        {
            (fileName[ii]) = '\0';
            break;  //'ENTER' key pressed
        }
        
        (fileName[ii]) = data;
        ii++; 
    }
    printf("entered string is = %s\r\n",fileName);
    sd_file_list();
    
    
    
}
void folder_files()
{
    ff_result = f_stat(foldername, &fno);
   
    switch (ff_result) 
    {
        case FR_OK:
                printf("FOLDER exist..\r\n");
                //create_file();
            break;

        case FR_NO_FILE:
            printf("FOLDER is not exist.\r\n");
            ff_result = f_mkdir(foldername);
            if (ff_result != FR_OK)
            {
              printf("Directory not created\r\n.");
            }
            else
            {
                printf("create DIR successful = %s\r\n",foldername);
            }
            break;

        default:
            printf("An error occured. (%d)\n", ff_result);
    }
}

void test()
{
    get_name(); 
    
    if(fileName[0]!= '/')
    {
        memcpy(&dirname[1],&fileName[0],(sizeof(fileName)-1));
        dirname[0] = '/';
    }
    else
    {
        memcpy(dirname,fileName,sizeof(fileName));
    }
    
    
    f_chdir(fileName);

}

int main(void)
{
    uint8_t option = 0;
    bsp_board_leds_init();
    uart_init();
    lfclk_config();
    rtc_config();
    fatfs_example();

    printf("UART Start..\r\n");
    
    while (true)
    {
        printf("Press any key...\r\n");
        while (app_uart_get(&option) != NRF_SUCCESS);
        printf("\r\n");
        printf("0: List Files\r\n");
        printf("1: Format SD card\t2:Check memory size \r\n");
        printf("3: Create New File\t4:Remove perticular File \r\n");
        printf("5: Append to file\t6:Read File\r\n");
        printf("7: Create Directory\t8:Read Current Timestamp\r\n");
        printf("9: Set TimeDate for RTC\ta:Read Current Time and Date from RTC\r\n");
        
        printf("\r\nSelect Any option between 1 to a\r\n");
        
        while (app_uart_get(&option) != NRF_SUCCESS);
        while (app_uart_put(option) != NRF_SUCCESS);
        switch (option)
        {
            case '0': printf("Listing the files...\r\n");
                      sd_file_list();
                      break;
            case '1': printf("\r\nFormatting SD card with FAT32...\r\n");
                      nrf_drv_rtc_disable(&rtc);
                      sd_format();
                      set_date_time();
                      break;
            case '2': printf("\r\nChecking File memory...\r\n");
                      sd_filesize();
                      break;
            case '3': printf("\r\nCreating New File...\r\n");
                      sd_create_file();
                      break;
            case '4': printf("\r\nRemoving File...\r\n");
                      sd_file_remove();
                      break;
            case '5': printf("\r\nAppending to file...\r\n");
                      sd_append_file();
                      break;
            case '6': printf("\r\nReading File...\r\n");
                      break;
            case '7': printf("\r\nCreating New directory...\r\n");
                      sd_create_dir();
                      break;
            case '8': printf("\r\nReading Timestamp...\r\n");
                      sd_time_date();
                      break;
            case '9': printf("\r\nSetting Start Date and Time for RTC...\r\n");
                      set_date_time();
                      break;
            case 'a': printf("\r\nGrtting Time and Date from RTC...\r\n");
                      get_time_date();
                      create_folder();
                      break;
            case 'b': printf("\r\nCreating files in folder...\r\n");
                      file_list();
                      break;
            case 'c': printf("\r\nReading The directory...\r\n");
                      read_dir();
                      break;
            case 'd': printf("\r\nChanging current directory..\r\n");
                      test();
                      break;                      
            default:   printf("\r\nDefault...\r\n");
        }
    }
}
